/**CFile***********************************************************************

  FileName    [bmcSat.c]

  PackageName [bmc]

  Synopsis    [The bmc-sat gateway implementation file]

  Description []

  SeeAlso     [the sat package]

  Author      [Roberto Cavada]

  Copyright   [
  This file is part of the ``bmc'' package of NuSMV version 2.
  Copyright (C) 2000-2001 by ITC-irst and University of Trento.

  NuSMV version 2 is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "bmcSat.h"
#include "bmcInt.h" /* for 'options' */

#include "be/be.h"
#include "sim/sim.h"
#include "mc/mc.h" /* for 'eval_spec' */
#include "parser/symbols.h" /* for constants */

#include "enc/enc.h"


/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/**Struct*********************************************************************

  Synopsis           [Private declaration for class Bmc_SatTrace]
  Description        []
  SeeAlso            []

******************************************************************************/
typedef struct Bmc_SatTrace_TAG {
  NodeList_ptr smvModel; /* the symbolic model */
  Sat_SatResult_ptr satResult; /* the original sover outcome */
  Bmc_Problem_ptr bmcProb; /* the original problem */
  int smvModel_k; /* the trace length (for lazy evaluation) */
  int cnfMaxVarIdx;  /* the maximum CNF variable index (for the explainer) */ 

} Bmc_SatTrace;


/**Struct*********************************************************************

  Synopsis           [Private declaration for class Bmc_Problem]
  Description        []
  SeeAlso            []

******************************************************************************/
typedef struct Bmc_Problem_TAG {
  be_ptr beProb;        /* the "original" problem, in be format */
  Be_Cnf_ptr cnfProb;   /* the corresponding CNF problem */
} Bmc_Problem;



/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static NodeList_ptr
bmc_sat_trace_prop2symb ARGS((const Bmc_SatTrace_ptr self, 
			      const BmcVarsMgr_ptr vars_mgr)); 

static void 
bmc_sat_trace_calculateSymbModel ARGS((const Bmc_SatTrace_ptr self,  
				       const BmcVarsMgr_ptr vars_mgr, 
				       const int k));

static void bmc_sat_trace_clear_model ARGS((Bmc_SatTrace_ptr self));

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Calls the specified solver with the given problem, 
  and returns a sat result if there is any]

  Description        [The returned result MUST be deleted by the 
  calling function.]

  SideEffects        [A Sat_SatResult_ptr will be created, 
  and the calling function must delete it.]

  SeeAlso            [Sat_SatSolver, Sat_SatResult_ptr]

******************************************************************************/
Sat_SatResult_ptr Bmc_Sat_SolveProblem(const BmcVarsMgr_ptr vars_mgr, 
				       Sat_SatSolver_ptr solver, 
				       const Bmc_Problem_ptr bmcProb, 
				       const int k)
{
  Sat_SatResult_ptr result = Sat_SatResult_Create();

  be_ptr beProb = Bmc_Problem_GetBeProblem(bmcProb);
  if (Be_IsTrue(BmcVarsMgr_GetBeMgr(vars_mgr), beProb)) {
    Sat_SatResult_SetSatisfiable(result);
    
    if (opt_verbose_level_gt(options, 0)) {    
      fprintf(nusmv_stderr, 
	 "The given problem has been evaluated as the constant value true.\n");
    }
  }
  else if (Be_IsFalse(BmcVarsMgr_GetBeMgr(vars_mgr), beProb)) {
    Sat_SatResult_SetUnsatisfiable(result);

    if (opt_verbose_level_gt(options, 0)) {    
      fprintf(nusmv_stdout, 
        "The given problem has been evaluated as the constant value false.\n");
    }
  }
  else {
    /*************************************************************
     * CALL SIM SOLVER INTERNALLY 
     * Using default parameters, which are hidden here
     *************************************************************/
    Sat_SatSolver_Solve(solver, Bmc_Problem_GetCnf(bmcProb), result);

    /*************************************************************
     * PROCESS SOLVER RESULTS
     *************************************************************/
    if (Sat_SatResult_IsError(result)) {
      fprintf( nusmv_stdout, "Solver: UNRECOVERABLE INTERNAL ERROR\n" );
    }
    else if (Sat_SatResult_IsUnsatisfiable(result)) {
      if (opt_verbose_level_gt(options, 0)) {
	fprintf( nusmv_stdout, "Solver: given problem is unsatisfiable\n" );
      }
    }
    else {
      /* problem here is satisfiable */
      nusmv_assert(Sat_SatResult_IsSatisfiable(result));
      if (opt_verbose_level_gt(options, 0)) {
	fprintf( nusmv_stdout, "Solver: given problem is satisfiable\n" );
      }
    }
  }

  return result;
}


/**Function********************************************************************

  Synopsis           [Calls the specified solver with the given invar problem, 
  and returns a sat result if there is any]

  Description        []

  SideEffects        [result will change]
  
  SeeAlso            []

******************************************************************************/
Sat_SatResult_ptr Bmc_Sat_SolveInvarProblem(const BmcVarsMgr_ptr vars_mgr, 
					    Sat_SatSolver_ptr solver, 
					    const Bmc_Problem_ptr bmcProb)
{
  return Bmc_Sat_SolveProblem(vars_mgr, solver, bmcProb, 1);
}



/* ====================================================================== */
/* Bmc_Problem class methods:                                             */
/* ====================================================================== */


/**Function********************************************************************

  Synopsis           [Class Bmc_Problem constructor]

  Description        []

  SideEffects        []
  
  SeeAlso            [Bmc_Problem_Delete]

******************************************************************************/
Bmc_Problem_ptr Bmc_Problem_Create(const BmcVarsMgr_ptr vars_mgr, 
				   const be_ptr prob)
{
  Bmc_Problem_ptr self = ALLOC(Bmc_Problem, 1);
  nusmv_assert(self != NULL);

  self->beProb = prob; /* keeps memory of the original be problem */

  /* performs the cnf conversion: */
  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, "\nConverting the BE problem into CNF problem...\n");
  }
  
  self->cnfProb = Be_ConvertToCnf(BmcVarsMgr_GetBeMgr(vars_mgr), 
				  self->beProb);

  if (opt_verbose_level_gt(options, 1)) {
    fprintf(nusmv_stderr, " Conversion returned maximum variable index = %d\n",
	    Be_Cnf_GetMaxVarIndex(self->cnfProb));
    fprintf(nusmv_stderr, " Length of list of clauses = %d\n", 
	    Be_Cnf_GetClausesNumber(self->cnfProb));
    fprintf(nusmv_stderr, " Length of list of variables = %d\n", 
	    Be_Cnf_GetVarsNumber(self->cnfProb));
  }
  
  return self;
}


/**Function********************************************************************

  Synopsis           [Bmc_Problem destructor]

  Description        [Call when you no longer need the Bmc_Problem instance.]

  SideEffects        []
  
  SeeAlso            [Bmc_Problem_Create]

******************************************************************************/
void Bmc_Problem_Delete(Bmc_Problem_ptr self)
{
  nusmv_assert(self != NULL);
  Be_Cnf_Delete(self->cnfProb);
  FREE(self);
}


/**Function********************************************************************

  Synopsis           [Accessor method to retrieve the original problem in 
  BE format]

  Description        [The returned be is the same you used during instance 
  construction, by passing it to Bmc_Problem_Create]

  SideEffects        []
  
  SeeAlso            []

******************************************************************************/
be_ptr Bmc_Problem_GetBeProblem(const Bmc_Problem_ptr self)
{
  return self->beProb;
}


/**Function********************************************************************

  Synopsis           [Returns the problem in CNF format]

  Description        [You do not need to delete the returned 
  Be_Cnf instance. If you delete the Bmc_Problem instance had returned 
  a Be_Cnf instance by calling this method, also the returned Be_Cnf instance 
  will become invalid.]

  SideEffects        []
  
  SeeAlso            []

******************************************************************************/
Be_Cnf_ptr Bmc_Problem_GetCnf(const Bmc_Problem_ptr self)
{
  return self->cnfProb;
}



/* ====================================================================== */
/* Bmc_SatTrace class methods:                                            */
/* ====================================================================== */

/**Function********************************************************************

  Synopsis           [Class Bmc_SatTrace constructor]

  Description        [The method gets a satisfiable Sat_SatResult instance.
  The ownership of both arguments bmcProb and satResult remanins to the 
  caller (i.e. the Bmc_SatTrace destroyer does not delete also these 
  instances.)]

  SideEffects        []
  
  SeeAlso            [Bmc_SatTrace_Delete]

******************************************************************************/
Bmc_SatTrace_ptr Bmc_SatTrace_Create(Bmc_Problem_ptr bmcProb, 
				     Sat_SatResult_ptr satResult)
{
  Bmc_SatTrace_ptr self = ALLOC(Bmc_SatTrace, 1);
  nusmv_assert(self != NULL);

  /* are you trying to call me with a non-satisfiable problem? */
  nusmv_assert(Sat_SatResult_IsSatisfiable(satResult)); 
  
  self->smvModel     = NodeList_create();
  self->satResult    = satResult;
  self->bmcProb      = bmcProb;
  self->smvModel_k   = 0;
  self->cnfMaxVarIdx = 0;
  
  return self;
}


/**Function********************************************************************

  Synopsis           [Class Bmc_SatTrace destructor]

  Description        [Call when you no longer need the given instance]

  SideEffects        [self will be invalidated]
  
  SeeAlso            []

******************************************************************************/
void Bmc_SatTrace_Delete(Bmc_SatTrace_ptr* self_ref) 
{
  bmc_sat_trace_clear_model(*self_ref);

  FREE(*self_ref);
  *self_ref = NULL;
}


/**Function********************************************************************

  Synopsis           [Returns the bdd-based symbolic model]

  Description        [Returns the symbolic counterexample with max length k. 
  The VarMgr instance is required by the conversion routine. 
  Conversion is performed with memoizing, so you can call efficiently call 
  this method more than one time. Returned object belongs to self]

  SideEffects        [self's internal state can change]
  
  SeeAlso            []

******************************************************************************/
NodeList_ptr Bmc_SatTrace_GetSymbModel(const Bmc_SatTrace_ptr self,  
				       const BmcVarsMgr_ptr vars_mgr, 
				       const int k)
{
  bmc_sat_trace_calculateSymbModel(self, vars_mgr, k);
  return self->smvModel;
}


/**Function********************************************************************

  Synopsis           [As Bmc_SatTrace_GetSymbModel, but for invariants]

  Description        [As Bmc_SatTrace_GetSymbModel, but for invariants. 
  Returned object belongs to self]

  SideEffects        []
  
  SeeAlso            []

******************************************************************************/
NodeList_ptr Bmc_SatTrace_GetSymbModelInvar(const Bmc_SatTrace_ptr self,  
					const BmcVarsMgr_ptr vars_mgr)
{
  bmc_sat_trace_calculateSymbModel(self, vars_mgr, 1);
  return self->smvModel;
}


/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Private service for class Bmc_SatTrace methods]

  Description        []

  SideEffects        []
  
  SeeAlso            []

******************************************************************************/
static void 
bmc_sat_trace_calculateSymbModel(const Bmc_SatTrace_ptr self,  
				 const BmcVarsMgr_ptr vars_mgr, 
				 const int k)
{
  be_ptr beProb = Bmc_Problem_GetBeProblem(self->bmcProb); 
  
  /* lazy evaluation on the original problem: */
  if (Be_IsTrue(BmcVarsMgr_GetBeMgr(vars_mgr), beProb) 
      || Be_IsFalse(BmcVarsMgr_GetBeMgr(vars_mgr), beProb)) {

    /* The counterexample explainer could be smarter (it might say "Is
       it a trivial problem?") */
    bmc_sat_trace_clear_model(self);
    self->smvModel = NodeList_create();
  }
  else {
    /* lazy evaluation on already performed conversions: */
    if ((NodeList_get_length(self->smvModel) == 0) || 
	(self->smvModel_k != k)) {
      self->smvModel_k = k;
      bmc_sat_trace_clear_model(self);
      self->smvModel = bmc_sat_trace_prop2symb(self, vars_mgr);
      self->cnfMaxVarIdx = 
	Be_Cnf_GetMaxVarIndex(Bmc_Problem_GetCnf(self->bmcProb));
    }  
  }
}



/* 
*/
/**Function********************************************************************

  Synopsis           [Private service for Bmc_SatTrace methods]

  Description        [Example of mapping of STATE VARIABLES and BE VARIABLES: 
  <PRE>
    ----------------------------------------------------------
    VAR        x    y    x'   y'   x0   y0   x1   y1   x2   y2
    BE index  0    1    2    3    4    5    6    7    8    9  
    CNF index  #    1    2    3    4    5    6    7    8    9 
    Time         -2    |   -1   |     0   |     1   |     2   |
    Varindex   0    1  | 0    1 |  0    1 |  0    1 |  0    1 |
    </PRE>
    ]

  SideEffects        []
  
  SeeAlso            []

******************************************************************************/
static NodeList_ptr
bmc_sat_trace_prop2symb(const Bmc_SatTrace_ptr self, 
			const BmcVarsMgr_ptr vars_mgr) 
{
  int state_input_vars_num;
  lsGen genLit;
  int lit;
  int** model; 
  int time;
  int varindex;
  BddEnc_ptr enc;
  NodeList_ptr trace; 

  state_input_vars_num = BmcVarsMgr_GetStateInputVarsNum(vars_mgr);

  trace = NodeList_create();

  if (opt_verbose_level_gt(options, 1)) {
    fprintf(nusmv_stderr, 
	    "The propositional to symbolic model converter says that "
	    "\n the input CNF model is the following:\n (");
    
    genLit = lsStart(Sat_SatResult_GetPropModel(self->satResult));
    while (lsNext(genLit, (lsGeneric*) &lit, LS_NH) == LS_OK) {
      fprintf(nusmv_stderr, "%d ", lit);
    }
    lsFinish(genLit);
    fprintf(nusmv_stderr, " )\n");
  }
  
  if (opt_verbose_level_gt(options, 0)) {
    fprintf( nusmv_stderr, 
	     "Converting propositional model into symbolic model\n" );
  }

  if (opt_verbose_level_gt(options, 1)) {
    fprintf(nusmv_stderr, 
	    "Number of state and input variables: %d, ", state_input_vars_num);
    fprintf(nusmv_stderr, "Time steps from 0 to %d\n", self->smvModel_k);
    fprintf(nusmv_stderr, "\nLiteral, model[time][var_index] and symbol:\n");
  }

  /* allocates memory for model, and initialize all locations (default
     value for any unassigned literal is 'false') */
  model = ALLOC(int*, (self->smvModel_k + 1));
  nusmv_assert(model != NULL);

  for(time = 0; time <= self->smvModel_k; ++time) {
    model[time] = ALLOC(int, state_input_vars_num);
    nusmv_assert(model[time] != NULL);
    /* resets the allocated array: */ 
    for (varindex = 0; varindex < state_input_vars_num; ++varindex) {
      model[time][varindex] = 0;
    }
  }

  /* scan list of literals */
  genLit = lsStart(Sat_SatResult_GetPropModel(self->satResult));
  while (lsNext(genLit, (lsGeneric*) &lit, LS_NH) == LS_OK) {
    int bevar_idx = abs(lit);

    /* here only state and input variables indexes are allowed: */
    if (BmcVarsMgr_IsBeIndexStateVar(vars_mgr, bevar_idx, self->smvModel_k) ||
	BmcVarsMgr_IsBeIndexInputVar(vars_mgr, bevar_idx, self->smvModel_k)) {

      time = BmcVarsMgr_BeIndex2Time(vars_mgr, bevar_idx); 
      varindex = BmcVarsMgr_BeIndex2VarIndex(vars_mgr, bevar_idx);
      model[time][varindex] = (lit > 0) ? 1 : 0;

      if (opt_verbose_level_gt(options, 1)) {
	fprintf(nusmv_stderr, "Literal %5d: ", lit);
	fprintf( nusmv_stderr, "model[%d][%d] = % d      ", time, varindex,
		 model[time][varindex] );
	print_node(nusmv_stderr, 
		   BmcVarsMgr_VarIndex2Name(vars_mgr, varindex));
	fprintf(nusmv_stderr, " == %d", model[time][varindex]);
	fprintf(nusmv_stderr, "\n");
      }
    }
  } /* end of literal scan */
  lsFinish(genLit);

  if (opt_verbose_level_gt(options, 1)) {
	fprintf(nusmv_stderr, "\n");
  }

  /*************************************************************
   * BUILD COUNTEREXAMPLE AS SEQUENCE OF BDDs
   *************************************************************/
  enc = Enc_get_bdd_encoding();
  BddEnc_print_bdd_begin(enc, 
			 NodeList_to_node_ptr(
				     Encoding_get_model_state_symbols_list(
					  BddEnc_get_symbolic_encoding(enc))), 
			 false);

  for (time = 0; time <= self->smvModel_k; ++time) {
    node_ptr state_exp = new_node(TRUEEXP, Nil, Nil);
    node_ptr input_exp = new_node(TRUEEXP, Nil, Nil);
    
    if (opt_verbose_level_gt(options, 10)) {
      fprintf(nusmv_stderr, " --------------- Time %d --------------- \n", 
	      time);
    }
    
    for (varindex = 0; varindex < state_input_vars_num; ++varindex) { 
      node_ptr var_expr;
      
      var_expr = BmcVarsMgr_VarIndex2Name(vars_mgr, varindex);
      
      /* negate if needed: */
      if(model[time][varindex] == 0) { 
	var_expr = new_node(NOT, var_expr, Nil);
      }

      if (BmcVarsMgr_IsBeIndexInCurrStateBlock(vars_mgr, varindex)) {
	state_exp = new_node(AND, var_expr, state_exp);
      }
      else {
	nusmv_assert(BmcVarsMgr_IsBeIndexInInputBlock(vars_mgr, varindex));
	input_exp = new_node(AND, var_expr, input_exp);
      }      
    } /* cycle on variables for a given time */
    
    /* Evaluates and append state and input to list: */
    {
      bdd_ptr state_bdd;
      bdd_ptr input_bdd;
	
      /* calculates, append and if needed print states */
      state_bdd = eval_spec( Prop_master_get_bdd_fsm(), enc, state_exp, Nil );
      NodeList_append(trace, (node_ptr) state_bdd);
      if (opt_verbose_level_gt(options, 10)) {
	printf("At time %d, STATES:\n", time);
	BddEnc_print_bdd(enc, state_bdd, nusmv_stdout);
      }

      /* at last step inputs are not appended */
      if (time != self->smvModel_k) {
	input_bdd = eval_spec( Prop_master_get_bdd_fsm(), enc, input_exp, Nil );
	NodeList_append(trace, (node_ptr) input_bdd); 
	if (opt_verbose_level_gt(options, 10)) {
	  BddEnc_print_bdd_begin(enc, 
		 NodeList_to_node_ptr(
			     Encoding_get_model_input_symbols_list(
				     BddEnc_get_symbolic_encoding(enc))), 
				 false);

	  printf("At time %d, INPUTS:\n", time);
	  BddEnc_print_bdd(enc, input_bdd, nusmv_stdout);
	  BddEnc_print_bdd_end(enc);
	}
      }

    }

  } /* end of time cycle */

  BddEnc_print_bdd_end(enc);

  /* destroys previously allocated model array: */
  for(time = 0; time <= self->smvModel_k; ++time) {
    FREE(model[time]);
  }
  FREE(model);

  return trace;
}



/**Function********************************************************************

  Synopsis           [Service function]

  Description        [Clears the list of bdds that represents the model]

  SideEffects        []
  
  SeeAlso            []

******************************************************************************/
static void bmc_sat_trace_clear_model(Bmc_SatTrace_ptr self) 
{
  ListIter_ptr iter;
  DdManager* dd = BddEnc_get_dd_manager(Enc_get_bdd_encoding());

  /* frees the list of bdds */
  iter = NodeList_get_first_iter(self->smvModel);
  while ( !ListIter_is_end(iter) ) {
    bdd_free(dd, (bdd_ptr) NodeList_get_value_at(self->smvModel, iter));

    iter = ListIter_get_next(iter);
  }
  NodeList_destroy(self->smvModel);  

  self->smvModel = NODE_LIST(NULL);
}
